unit Unit_Server_Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ScktComp, WinSock, Grids, ComCtrls, StdCtrls, ExtCtrls;

  resourcestring
    StatusDateTimeFormat = 'mm/dd/yyyy" - "hh:nn:ss:zzz AM/PM';
    WelcomeMessage = 'Welcome to Simple Chat Server 1.0!' + #13#10 +
                     '----------------------------------' + #13#10;
    ConnectHlpMsg = 'Commands: Help(?)' + #13#10 +
                    '          Refresh chat list(")' + #13#10 +
                    '          Set ScreenName(SNAME=<Screen Name>)' + #13#10;

  type TServerConnectionStatus = record
    tcsActiveConnections : Integer;
    tcsScreenName        : string;
    tcsLocalHost         : string;
    tcsLocalAddress      : string;
    tcsLocalPort         : Integer;
    tcsRemoteHost        : string;
    tcsRemoteAddress     : string;
    tcsRemotePort        : Integer;
    tcsServerPrivateMsg  : Boolean;
    tcsSocketHandle      : Integer;
    tcsServerSocketHWND  : HWND;
  end;

  type TServerConnectionStatusArray = array of TServerConnectionStatus;

  type
    TForm_Server_Main = class(TForm)
      ChatServerSocket: TServerSocket;
      pcLearnSockets: TPageControl;
      tsServer: TTabSheet;
      gbServerSettings: TGroupBox;
      lblServerPort: TLabel;
      Label2: TLabel;
      lblServerThreadCacheSize: TLabel;
      btnServerOpen: TButton;
      btnServerClose: TButton;
      edServerPort: TEdit;
      edServerService: TEdit;
      edServerThreadCacheSize: TEdit;
      gbServerSocketType: TGroupBox;
      rbtnServerNonBlocking: TRadioButton;
      rdbtnServerThreadBlocking: TRadioButton;
      gbUserDefinedServerSettings: TGroupBox;
      TabSheet1: TTabSheet;
      TabSheet2: TTabSheet;
      lvServerActivity: TListView;
      btnClearServerSocketLog: TButton;
      sgServerConnections: TStringGrid;
      memReceive: TMemo;
      edSend: TEdit;
      cbBroadcastMessage: TCheckBox;
    cbLogAllClientMessages: TCheckBox;
      procedure btnServerOpenClick(Sender: TObject);
      procedure btnServerCloseClick(Sender: TObject);
      procedure ChatServerSocketAccept(Sender: TObject; Socket: TCustomWinSocket);
      procedure ChatServerSocketClientConnect(Sender: TObject; Socket: TCustomWinSocket);
      procedure ChatServerSocketClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
      procedure ChatServerSocketClientError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
      procedure ChatServerSocketClientRead(Sender: TObject; Socket: TCustomWinSocket);
      procedure ChatServerSocketGetSocket(Sender: TObject; Socket: Integer; var ClientSocket: TServerClientWinSocket);
      procedure ChatServerSocketClientWrite(Sender: TObject; Socket: TCustomWinSocket);
      procedure ChatServerSocketGetThread(Sender: TObject; ClientSocket: TServerClientWinSocket; var SocketThread: TServerClientThread);
      procedure ChatServerSocketListen(Sender: TObject; Socket: TCustomWinSocket);
      procedure ChatServerSocketThreadEnd(Sender: TObject; Thread: TServerClientThread);
      procedure ChatServerSocketThreadStart(Sender: TObject; Thread: TServerClientThread);
      procedure btnClearServerSocketLogClick(Sender: TObject);
      procedure edSendKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
      procedure sgServerConnectionsDblClick(Sender: TObject);
      procedure sgServerConnectionsKeyDown(Sender: TObject; var Key: Word;Shift: TShiftState);
      procedure FormCreate(Sender: TObject);
      procedure FormClose(Sender: TObject; var Action: TCloseAction);
      procedure memReceiveKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
      //
    private
      { Private declarations }
      procedure AddConnectionsInfo(ClientInfo : TServerConnectionStatus);
      procedure DeleteConnectionsInfo(ClientInfo : TServerConnectionStatus); overload;
      procedure DeleteConnectionsInfo(I : Integer); overload;
      procedure DisplayConnectionsInfo;
      procedure UpdateLoginInfo(const Socket: TCustomWinSocket; const Data : string);
      procedure ServerBrodcastMessage;
      function  CreateLoginList : string;
      procedure BroadcastLoginListUpdate(ExcludeSocket: TCustomWinSocket);
      procedure ProxyMessage(ReadText : string; SenderSocket : TCustomWinSocket);
    public
      { Public declarations }
  end;

  var
    Form_Server_Main: TForm_Server_Main;
    ServerConnectionStatus : TServerConnectionStatusArray;

implementation

{$R *.DFM}

procedure TForm_Server_Main.btnServerOpenClick(Sender: TObject);
var
  PrevCursor : TCursor;
  X : TListItem;
begin
  PrevCursor := Screen.Cursor;
  Screen.Cursor := crHourglass;

  DisplayConnectionsInfo;
  ChatServerSocket.ThreadCacheSize := StrToInt(edServerThreadCacheSize.Text);
  ChatServerSocket.Port := StrToInt(edServerPort.Text);
//  if (rbtnServerNonBlocking.Checked) then
    ChatServerSocket.ServerType := stNonBlocking;
//  if (rdbtnServerThreadBlocking.Checked) then
//    ChatServerSocket.ServerType := stThreadBlocking;
  X := lvServerActivity.Items.Add;
  X.Caption := 'Opening...';
  ChatServerSocket.Open;
  X.Caption := 'Opened';
  memReceive.Lines.Append('Started...' + #13#10);
  X.SubItems.Add(ChatServerSocket.Socket.LocalHost);
  X.SubItems.Add(ChatServerSocket.Socket.LocalAddress);
  X.SubItems.Add(IntToStr(ChatServerSocket.Socket.LocalPort));

  Screen.Cursor := PrevCursor;
end;

procedure TForm_Server_Main.btnServerCloseClick(Sender: TObject);
var
  I : Integer;
  X : TListItem;
begin
  SetLength(ServerConnectionStatus, 0);
  ServerConnectionStatus := nil;
  X := lvServerActivity.Items.Add;
  X.Caption := 'Closing...';
  X.SubItems.Add(ChatServerSocket.Socket.LocalHost);
  X.SubItems.Add(ChatServerSocket.Socket.LocalAddress);
  X.SubItems.Add(IntToStr(ChatServerSocket.Socket.LocalPort));
  with ChatServerSocket.Socket do
  begin
    for I := 0 to (ActiveConnections - 1) do
      if Connections[I].Connected then Connections[I].Close;
  end;
  ChatServerSocket.Close;
  X.Caption := 'Closed';
  memReceive.Lines.Append('...Closed' + #13#10)
end;

procedure TForm_Server_Main.ChatServerSocketClientConnect(Sender: TObject; Socket: TCustomWinSocket);
var
  ClientConnectInfo : TServerConnectionStatus;
  X : TListItem;
begin
  X := lvServerActivity.Items.Add;
  X.Caption := 'Client Connecting...';
  with ClientConnectInfo do
  begin
    tcsActiveConnections := -1;
    tcsScreenName        := 'Anonymous';
    tcsLocalHost         := Socket.LocalHost;
    tcsLocalAddress      := Socket.LocalAddress;
    tcsLocalPort         := Socket.LocalPort;
    tcsRemoteHost        := Socket.RemoteHost;
    tcsRemoteAddress     := Socket.RemoteAddress;
    tcsRemotePort        := Socket.RemotePort;
    tcsServerPrivateMsg  := False;
    tcsSocketHandle      := Socket.SocketHandle;
    tcsServerSocketHWND  := Socket.Handle;
  end;
  AddConnectionsInfo(ClientConnectInfo);
  DisplayConnectionsInfo;
  X.Caption := 'Client Connected ';
  X.SubItems.Add(Socket.RemoteHost);
  X.SubItems.Add(Socket.RemoteAddress);
  X.SubItems.Add(IntToStr(Socket.RemotePort));
  X.SubItems.Add(FormatDateTime(StatusDateTimeFormat, Now));
  Socket.SendText(WelcomeMessage +
                  'Date/Time: ' + FormatDateTime(StatusDateTimeFormat, Now) + #13#10 +
                  ConnectHlpMsg + #13#10);
end;

procedure TForm_Server_Main.ChatServerSocketClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
var
  ClientConnectInfo : TServerConnectionStatus;
  X : TListItem;
begin
  X := lvServerActivity.Items.Add;
  X.Caption := 'Client Disconnecting...';
  X.SubItems.Add(Socket.RemoteHost);
  X.SubItems.Add(Socket.RemoteAddress);
  X.SubItems.Add(IntToStr(Socket.RemotePort));
  with ClientConnectInfo do
  begin
    tcsActiveConnections := -1;
    tcsScreenName        := '';
    tcsLocalHost         := Socket.LocalHost;
    tcsLocalAddress      := Socket.LocalAddress;
    tcsLocalPort         := Socket.LocalPort;
    tcsRemoteHost        := Socket.RemoteHost;
    tcsRemoteAddress     := Socket.RemoteAddress;
    tcsRemotePort        := Socket.RemotePort;
    tcsSocketHandle      := Socket.SocketHandle;
    tcsServerSocketHWND  := Socket.Handle;
    tcsServerPrivateMsg  := False;
  end;
  DeleteConnectionsInfo(ClientConnectInfo);
  DisplayConnectionsInfo;
  X.Caption := 'Client Disconnected';
  X.SubItems.Add(FormatDateTime(StatusDateTimeFormat, Now));
  BroadcastLoginListUpdate(Socket);
end;

procedure TForm_Server_Main.ChatServerSocketClientError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
var
  X : TListItem;
begin
  X := lvServerActivity.Items.Add;
  X.Caption := 'Client Error: #' + IntToStr(ErrorCode);
  X.SubItems.Add(Socket.RemoteHost);
  X.SubItems.Add(Socket.RemoteAddress);
  X.SubItems.Add(IntToStr(Socket.RemotePort));
//ErrorExents
//eeGeneral	   //The socket received an error message that does not fit into any of the following categories.
//eeSend	     //An error occurred when trying to write to the socket connection.
//eeReceive	   //An error occurred when trying to read from the socket connection.
//eeConnect	   //A connection request that was already accepted could not be completed.
//eeDisconnect //An error occurred when trying to close a connection.
//eeAccept	   //A problem occurred when trying to accept a client connection request.
end;

procedure TForm_Server_Main.ChatServerSocketClientRead(Sender: TObject; Socket: TCustomWinSocket);
var
  X : TListItem;
  ReadText : string;
begin
  ReadText := Socket.ReceiveText;
  if (cbLogAllClientMessages.Checked) then
  begin
    X := lvServerActivity.Items.Add;
    X.Caption := 'Client Read ';
    X.SubItems.Add(Socket.RemoteHost);
    X.SubItems.Add(Socket.RemoteAddress);
    X.SubItems.Add(IntToStr(Socket.RemotePort));
    X.SubItems.Add(FormatDateTime(StatusDateTimeFormat, Now));
    X.SubItems.Add(ReadText);
  end;
  if (UpperCase(Copy(ReadText, 1 ,6)) = 'SNAME=') then
    begin
      UpdateLoginInfo(Socket, ReadText);
      Socket.SendText(CreateLoginList);
      BroadcastLoginListUpdate(Socket);
    end
  else if (Copy(ReadText, 1 ,1) = '"') then //Request for login list
    Socket.SendText(CreateLoginList)
  else if (Copy(ReadText, 1 ,1) = '?') then //Request for help
    Socket.SendText('Date/Time: ' + FormatDateTime(StatusDateTimeFormat, Now) + #13#10 + ConnectHlpMsg)
  else if (Copy(ReadText, 1 ,1) = '<') then //Client Sending Message
    ProxyMessage(ReadText, Socket) //parse and redirect
  else //Garbage Input Received so drop connection
    Socket.Close;
end;

procedure TForm_Server_Main.ChatServerSocketClientWrite(Sender: TObject; Socket: TCustomWinSocket);
var
  X : TListItem;
begin
  X := lvServerActivity.Items.Add;
  X.Caption := 'Client Write ';
  X.SubItems.Add(Socket.RemoteHost);
  X.SubItems.Add(Socket.RemoteAddress);
  X.SubItems.Add(IntToStr(Socket.RemotePort));
  X.SubItems.Add(FormatDateTime(StatusDateTimeFormat, Now));
end;

procedure TForm_Server_Main.ChatServerSocketListen(Sender: TObject; Socket: TCustomWinSocket);
var
  X : TListItem;
begin
  X := lvServerActivity.Items.Add;
  X.Caption := 'Listening...';
  X.SubItems.Add(Socket.LocalHost);
  X.SubItems.Add(Socket.LocalAddress);
  X.SubItems.Add(IntToStr(Socket.LocalPort));
  X.SubItems.Add(FormatDateTime(StatusDateTimeFormat, Now));
end;

procedure TForm_Server_Main.btnClearServerSocketLogClick(Sender: TObject);
begin
  DisplayConnectionsInfo;
  lvServerActivity.Items.Clear;
end;

procedure TForm_Server_Main.edSendKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  I, J : Integer;
begin
  if Key = VK_RETURN then
  begin
    if (ChatServerSocket.Active) and (cbBroadcastMessage.Checked) then
      ServerBrodcastMessage
    else if (ChatServerSocket.Active) then //Server specific message
      for I := 0 to High(ServerConnectionStatus) do
      begin
        //Send message only to those selected
        if (ServerConnectionStatus[I].tcsServerPrivateMsg) then
        begin
          //Find the user to send message
          for J := 0 to (ChatServerSocket.Socket.ActiveConnections - 1) do
          begin
            with ChatServerSocket.Socket.Connections[J] do
            begin
              if (ServerConnectionStatus[I].tcsRemoteAddress = RemoteAddress) and
               (ServerConnectionStatus[I].tcsRemotePort = RemotePort) then
               begin
                 if (Connected) then SendText('Server Admin: ' + edSend.Text + #13#10);
                 Break;
               end
            end;
          end;
        end;
      end;
    memReceive.Lines.Append('Server Admin: ' + edSend.Text + #13#10);
    edSend.Text := '';
  end;
end;

procedure TForm_Server_Main.sgServerConnectionsDblClick(Sender: TObject);
var
  I : Integer;
begin
  I := sgServerConnections.Row - 1;
  if (sgServerConnections.Row <> 0) then
    with ServerConnectionStatus[I] do
    begin
      if (tcsServerPrivateMsg) then
        begin
          tcsServerPrivateMsg := False;
          sgServerConnections.Cells[0,sgServerConnections.Row] := '';
        end
      else
        begin
          tcsServerPrivateMsg := True;
          sgServerConnections.Cells[0,sgServerConnections.Row] := '(Selected)';
        end;
    end;
end;

procedure TForm_Server_Main.ChatServerSocketThreadEnd(Sender: TObject; Thread: TServerClientThread);
begin
//
end;

procedure TForm_Server_Main.ChatServerSocketThreadStart(Sender: TObject; Thread: TServerClientThread);
begin
//
end;

procedure TForm_Server_Main.ChatServerSocketGetSocket(Sender: TObject; Socket: Integer; var ClientSocket: TServerClientWinSocket);
begin
//
end;

procedure TForm_Server_Main.ChatServerSocketGetThread(Sender: TObject; ClientSocket: TServerClientWinSocket; var SocketThread: TServerClientThread);
begin
//
end;

procedure TForm_Server_Main.ChatServerSocketAccept(Sender: TObject; Socket: TCustomWinSocket);
begin
//
end;

procedure TForm_Server_Main.DisplayConnectionsInfo;
var
  I : Integer;
begin
  with ChatServerSocket do
  begin
    //Setup the grid
    //Setup RowCount
    if (Length(ServerConnectionStatus) >= 1) then
      sgServerConnections.RowCount := Length(ServerConnectionStatus) + 1
    else
      sgServerConnections.RowCount := 2;
    //Setup Fixed Rows
    sgServerConnections.FixedRows := 1;
    //Setup Columns
    with sgServerConnections do
    begin
      //Fill grid with data
      for I := 0 to High(ServerConnectionStatus) do
      begin
        with ServerConnectionStatus[I] do
        begin
          Cells[1 ,I+1] := IntToStr(tcsActiveConnections);
          Cells[2 ,I+1] := tcsScreenName;
          Cells[3 ,I+1] := tcsLocalHost;
          Cells[4 ,I+1] := tcsLocalAddress;
          Cells[5 ,I+1] := IntToStr(tcsLocalPort);
          Cells[6 ,I+1] := tcsRemoteHost;
          Cells[7 ,I+1] := tcsRemoteAddress;
          Cells[8 ,I+1] := IntToStr(tcsRemotePort);
          if (tcsServerPrivateMsg) then
            Cells[0 ,I+1] := '(Selected)'
          else
            Cells[0 ,I+1] := '';
        end;
      end; //end for I
    end; //end with sgServerConnections
  end; //end with ServerSocket
end;

procedure TForm_Server_Main.UpdateLoginInfo(const Socket: TCustomWinSocket; const Data : string);
var
  I : Integer;
begin
  for I := 0 to High(ServerConnectionStatus) do
    with ServerConnectionStatus[I] do
    begin
      if (tcsRemoteAddress = Socket.RemoteAddress) and
         (tcsRemotePort = Socket.RemotePort) then
         begin
           tcsScreenName := Copy(Data, 7 ,Length(Data) - 5);
           Break;
         end;
    end;
  DisplayConnectionsInfo;
end;

procedure TForm_Server_Main.ServerBrodcastMessage;
var
  I : Integer;
begin
  with ChatServerSocket.Socket do
  begin
    for I := 0 to (ActiveConnections - 1) do
      Connections[I].SendText('Server Admin: ' + edSend.Text + #13#10);
    memReceive.Lines.Append('Server Admin: ' + edSend.Text + #13#10);
  end;
end;

procedure TForm_Server_Main.AddConnectionsInfo(ClientInfo : TServerConnectionStatus);
var
  I, X     : Integer;
  AddToSCS : Boolean;
begin
  //Search for an empty slot in the ServerConnectionStatus array
  AddToSCS := True;
  for I := 0 to High(ServerConnectionStatus) do
  begin
    if (ServerConnectionStatus[I].tcsScreenName = 'Empty') then
      begin
        AddToSCS := False;
        Break;
      end;
  end;

  if (AddToSCS) then //An empty slot was not found so add one and use it
    begin
      SetLength(ServerConnectionStatus, Length(ServerConnectionStatus) + 1);
      X := High(ServerConnectionStatus);
    end
  else
    X := I;
  //Add the new connection information
  with ServerConnectionStatus[X] do
  begin
    tcsActiveConnections := X;
    tcsScreenName        := ClientInfo.tcsScreenName;
    tcsLocalHost         := ClientInfo.tcsLocalHost;
    tcsLocalAddress      := ClientInfo.tcsLocalAddress;
    tcsLocalPort         := ClientInfo.tcsLocalPort;
    tcsRemoteHost        := ClientInfo.tcsRemoteHost;
    tcsRemoteAddress     := ClientInfo.tcsRemoteAddress;
    tcsRemotePort        := ClientInfo.tcsRemotePort;
    tcsServerPrivateMsg  := False;
    tcsSocketHandle      := ClientInfo.tcsSocketHandle;
    tcsServerSocketHWND  := ClientInfo.tcsServerSocketHWND;
  end;
end;

procedure TForm_Server_Main.DeleteConnectionsInfo(ClientInfo : TServerConnectionStatus);
var
  J : Integer;
begin
  //Find the Entry and clear it
  for J := 0 to High(ServerConnectionStatus) do
    with ServerConnectionStatus[J] do
    begin
      if (tcsRemoteAddress = ClientInfo.tcsRemoteAddress) and
         (tcsRemotePort = ClientInfo.tcsRemotePort) then
         begin
           tcsActiveConnections := 0;
           tcsScreenName        := 'Empty';
           tcsLocalHost         := '';
           tcsLocalAddress      := '';
           tcsLocalPort         := 0;
           tcsRemoteHost        := '';
           tcsRemoteAddress     := '';
           tcsRemotePort        := 0;
           tcsServerPrivateMsg  := False;
           tcsSocketHandle      := INVALID_SOCKET;
           tcsServerSocketHWND  := 0;
           Break;
         end;
    end;
end;

procedure TForm_Server_Main.DeleteConnectionsInfo(I : Integer);
begin
  with ServerConnectionStatus[I] do
  begin
     tcsActiveConnections := 0;
     tcsScreenName        := 'Empty';
     tcsLocalHost         := '';
     tcsLocalAddress      := '';
     tcsLocalPort         := 0;
     tcsRemoteHost        := '';
     tcsRemoteAddress     := '';
     tcsRemotePort        := 0;
     tcsServerPrivateMsg  := False;
     tcsSocketHandle      := INVALID_SOCKET;
     tcsServerSocketHWND  := 0;
   end;
end;

function TForm_Server_Main.CreateLoginList : string;
var
  I : Integer;
begin
  Result := '"Server Admin"';
  for I := 0 to High(ServerConnectionStatus) do
    if (ServerConnectionStatus[I].tcsScreenName <> 'Empty') then
      Result := Result + ',"' + ServerConnectionStatus[I].tcsScreenName + '"';
end;

procedure TForm_Server_Main.BroadcastLoginListUpdate(ExcludeSocket: TCustomWinSocket);
var
  I : Integer;
  //LoginList : string;
begin
  //LoginList := CreateLoginList;
  for I := 0 to (ChatServerSocket.Socket.ActiveConnections - 1) do
    with ChatServerSocket.Socket.Connections[I] do
    begin
      if ((ExcludeSocket.RemoteAddress <> RemoteAddress) or
          (ExcludeSocket.RemotePort <> RemotePort)) then
          if Connected then SendText('*'); //Let the other users know that their list should be refreshed
          //if Connected then SendText(LoginList); //this could be used to update their list automatically
    end;
end;

procedure TForm_Server_Main.sgServerConnectionsKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  I, J : Integer;
  X : TListItem;
begin
  if ((Key = VK_DELETE) and (sgServerConnections.Row <> 0)) then
    begin
      I := sgServerConnections.Row - 1;
      for J := 0 to (ChatServerSocket.Socket.ActiveConnections - 1) do
        with ChatServerSocket.Socket.Connections[J] do
        begin
          if (ServerConnectionStatus[I].tcsRemoteAddress = RemoteAddress) and
             (ServerConnectionStatus[I].tcsRemotePort = RemotePort) then
             if Connected then
             begin
               X := lvServerActivity.Items.Add;
               X.Caption := 'Client Disconnecting...';
               X.SubItems.Add(RemoteHost);
               X.SubItems.Add(RemoteAddress);
               X.SubItems.Add(IntToStr(RemotePort));
               Close;
               X.Caption := 'Client Disconnected';
               X.SubItems.Add(FormatDateTime(StatusDateTimeFormat, Now));
               Break;
            end;
        end;
      DeleteConnectionsInfo(I);
      DisplayConnectionsInfo;
    end;
end;

procedure TForm_Server_Main.FormCreate(Sender: TObject);
begin
  //Setup Columns
  with sgServerConnections do
  begin
    Cells[1,0] := '#';
    ColWidths[1] := 25;
    Cells[2,0] := 'Screen Name';
    ColWidths[2] := 100;
    Cells[3,0] := 'Local Host';
    ColWidths[3] := 100;
    Cells[4,0] := 'Local Address';
    ColWidths[4] := 100;
    Cells[5,0] := 'Local Port';
    ColWidths[5] := 75;
    Cells[6,0] := 'Remote Host';
    ColWidths[6] := 100;
    Cells[7,0] := 'Remote Address';
    ColWidths[7] := 100;
    Cells[8,0] := 'Remote Port';
    ColWidths[8] := 75;
  end;
end;

procedure TForm_Server_Main.FormClose(Sender: TObject; var Action: TCloseAction);
var
  I : Integer;
begin
  with ChatServerSocket.Socket do
  begin
    for I := 0 to (ActiveConnections - 1) do
      if Connections[I].Connected then Connections[I].Close;
  end;
end;

procedure TForm_Server_Main.memReceiveKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key = VK_DELETE) then memReceive.Clear;
end;

procedure TForm_Server_Main.ProxyMessage(ReadText : string; SenderSocket : TCustomWinSocket);
var
  I, J, K, Y : Integer;
  X : string[1];
  SenderBegin, SenderLen,
    ReceiversBegin, ReceviersLen,
    MsgBegin, MsgLength : Integer;
  Sender : string;
  Receivers : TStringList;
  Msg : string;
  Found : Boolean;
begin
  Y := 0;
  SenderBegin := Y;
  SenderLen := Y;
  ReceiversBegin := Y;
  ReceviersLen := Y;
  MsgBegin := Y;
  MsgLength := Y;
  for I := 1 to Length(ReadText) do
  begin
    X := Copy(ReadText, I, 1);
    if (X = '<') then
      begin
        Inc(Y);
        if (Y = 1) then //Sender Begin
          SenderBegin := I + 1
        else if (Y = 2) then //Receivers Begin
          ReceiversBegin := I + 1
        else if (Y = 3) then //Message Begin
          MsgBegin := I + 1;
      end
    else if (X = '>') then
      begin
        if (Y = 1) then
          SenderLen := I - SenderBegin
        else if (Y = 2) then
          ReceviersLen := I - ReceiversBegin
        else if (Y = 3) then
          MsgLength := I - MsgBegin;
      end;
  end;
  //Copy data
  Sender := Copy(ReadText, SenderBegin, SenderLen);
  Receivers := TStringList.Create;
  Receivers.CommaText := Copy(ReadText, ReceiversBegin, ReceviersLen);
  Msg := Copy(ReadText, MsgBegin, MsgLength);

  //Debug Messages <"Server Admin","Anonymous","Anonymous2"><Hi>
  //memReceive.Lines.Append('---Debug Transmission---');
  //memReceive.Lines.Append('   Sender(' + IntToStr(SenderBegin) + ',' + IntToStr(SenderLen) + ') [' + Sender + ']' + #13#10 +
  //                        '   Receivers(' + IntToStr(ReceiversBegin) + ',' + IntToStr(ReceviersLen) + ') [' + Receivers.CommaText + ']' +  #13#10 +
  //                        '   Msg(' + IntToStr(MsgBegin) + ',' + IntToStr(MsgLength) + ') [' + Msg + ']');
  //memReceive.Lines.Append('---Debug Messaging---');

  if (Receivers.Count <= 0) then
    SenderSocket.SendText('Server: At least 1 receiver must be specified' + #13#10 +
                          '        in order to send a message' + #13#10);
  //Find receivers and send message
  for I := 0 to Receivers.Count - 1 do
  begin
    //Debug receivers
    //memReceive.Lines.Append('  *Receiver: ' + IntToStr(I) + '[' + Receivers[I] + ']');
    //Locate receivers in ServerConnectionStatus
    Found := False;
    for J := 0 to High(ServerConnectionStatus) do
      with ServerConnectionStatus[J] do
      begin
        if (tcsScreenName = Receivers[I]) then
        begin
          //Debug receivers found in list
          //memReceive.Lines.Append('     User: [' + Receivers[I] + '] Found');
          Found := True;
          Break;
        end;
      end;
    //Find socket and send message
    if Found then
      for K := 0 to (ChatServerSocket.Socket.ActiveConnections - 1) do
      with ChatServerSocket.Socket.Connections[K] do
      begin
        if ((ServerConnectionStatus[J].tcsRemoteAddress = RemoteAddress) and
           (ServerConnectionStatus[J].tcsRemotePort = RemotePort) and
           (Connected)) then
            begin
              //Debug receiver scoket found
              //memReceive.Lines.Append('     Socket: [' + Receivers[I] + '] Found');
              SendText(Sender + ': ' + Msg);
              Break;
            end;
      end
    else
      begin
//        if (cbLogAllClientMessages.Checked) then memReceive.Lines.Append(ReadText);
        if (Receivers[I] = 'Server Admin') then
          memReceive.Lines.Append(Sender + ': ' + Msg)
        else
          SenderSocket.SendText('Server: Unable to send message to ' + Receivers[I] + #13#10);
      end;
  end;
  //memReceive.Lines.Append('---------------------' + #13#10);
end;

end.
